(* ::Package:: *)
(*AtomicDensityMatrix`Common`*)

BeginPackage["AtomicDensityMatrix`Common`",{"AtomicDensityMatrix`UsageMessages`"}]

(* ::Subsection:: *)
(*Usage messages*)

(* ::Subsubsection:: *)
(*Atomic system*)

S::usage=""

L::usage=""

J::usage=""

NuclearSpin::usage=""

Parity::usage=""

Energy::usage=""

NaturalWidth::usage=""

HyperfineA::usage=""

HyperfineB::usage=""

HyperfineC::usage=""

Even::usage=""

Odd::usage=""

PolarizationMoments::usage=""

(* ::Subsubsection:: *)
(*Hamiltonians*)

ElectricField::usage=""

MagneticField::usage=""

(* ::Subsubsection:: *)
(*Math functions*)

HalfIntegerQ::usage="HalfIntegerQ[expr] gives True if expr is integer or half-integer, and False otherwise."

BlockDiagonal::usage="BlockDiagonal[matlist] forms a block diagonal matrix from a list of matrices matlist."

Commutator::usage = "Commutator[mat1, mat2] gives the commutator of matrices mat1 and mat2."

Anticommutator::usage = "Anticommutator[mat1, mat2] gives the anticommutator of matrices mat1 and mat2."

LowerTriangle::usage = "LowerTriangle[mat] gives a list of the elements in the lower triangle of mat, including the diagonal."

UpperTriangle::usage = "UpperTriangle[mat] gives a list of the elements in the upper triangle of mat, including the diagonal."

DirectProduct::usage = ""

DirectSelfProduct::usage = ""

EnsureList::usage = ""

AllSameQ::usage = ""

Slope::usage=""

ReplaceAt::usage=
    "ReplaceAt[expr, rules, pos] applies a rule or list of rules in an attempt to transform the part of expression expr specified by pos. 
ReplaceAt[expr, rules, pos, levelspec] applies rules at part of expression expr specified by pos and below according to the level specification levelspec relative to the position pos. 
ReplaceAt uses the default options of Replace.";

ReplaceAllAt::usage=
    "ReplaceAllAt[expr, rules, pos] applies a rule or list of rules in an attempt to transform all parts of the part of expression expr specified by pos."

(* The following are not used within the package 

FullSimplifySides::usage=""

SimplifySides::usage=""

Side::usage = ""

Both::usage = ""

DSolveIteration::usage = ""

DSolveIterated::usage = ""

LowestOrder::usage = ""

LowestOrderTerm::usage = ""

ReIm::usage = ""

ToEulerAngles::usage = "ToEulerAngles[vec] returns the Euler angles of the vector vec specified in cartesian coordinates."

AxisRotationMatrix::usage = ""

MaxPosition::usage = ""

EmbedMatrix::usage=""

FirstSeriesTerm::usage=""

PerturbativeIterate::usage=""

PerturbativeSolve::usage=""

IterationFunction::usage=""

PrintReturn::usage=""

*)

(* ::Subsection:: *)
(*Error messages*)

(* ::Section:: *)
(*Implementation*)

(* ::Subsection:: *)
(*Begin private context*)

Begin["`Private`"]

(* ::Subsection:: *)
(*Math functions*)

HalfIntegerQ[x_] := IntegerQ[2x]

Commutator[A_,B_]:=A.B-B.A

Anticommutator[A_,B_]:=(A.B+B.A)

LowerTriangle[matrix_?MatrixQ]:=Flatten[MapIndexed[Take[#1,#2[[1]]]&,matrix]]

UpperTriangle[matrix_?MatrixQ]:=LowerTriangle[Transpose@matrix]

DirectProduct[f_,T1_List,T2_List]:=Outer[f,T1,T2,1]

DirectProduct[T1_List,T2_List]:=DirectProduct[Times,T1,T2]

DirectSelfProduct[f_,T1_List]:=DirectProduct[f,T1,T1]

DirectSelfProduct[T1_List]:=DirectProduct[T1,T1]

BlockDiagonal[m:{__?MatrixQ}] :=
  Module[ {d = Dimensions/@m},
    ArrayFlatten@Table[
    	If[ i===j, m[[i]], Table[0,{d[[i,1]]},{d[[j,2]]}]],
    	{i,Length[m]},{j,Length[m]}
    ]
  ]

Slope[{{x1_,y1_},{x2_,y2_}}]:=If[x2!=x1,(y2-y1)/(x2-x1),Infinity]

EnsureList[expr_]:=If[Head[expr]===List,expr,{expr}];

AllSameQ[list_, expr_] := And @@ (expr === # & /@ list)

ReplaceAt[expr_, 
   rules : ((_Rule | _RuleDelayed) | {(_Rule | _RuleDelayed) ...}), 
   pos : {}, 
   levelspec : (_Integer | 
       Infinity | {_Integer | 
         Infinity} | {(_Integer | Infinity), (_Integer | 
          Infinity)}) : {0}, opts : OptionsPattern[]] := 
  Replace[expr, Flatten[{rules}], levelspec, opts];

ReplaceAt[expr_, 
   rules : ((_Rule | _RuleDelayed) | {(_Rule | _RuleDelayed) ...}), 
   pos : ({__Integer} | _Integer), 
   levelspec : (_Integer | 
       Infinity | {_Integer | 
         Infinity} | {(_Integer | Infinity), (_Integer | 
          Infinity)}) : {0}, opts : OptionsPattern[]] := 
  MapAt[Replace[#, Flatten[{rules}], levelspec, opts] &, expr, 
   Flatten[{pos}]];

ReplaceAt[expr_, 
   rules : ((_Rule | _RuleDelayed) | {(_Rule | _RuleDelayed) ...}), 
   pos : {__List}, 
   levelspec : (_Integer | 
       Infinity | {_Integer | 
         Infinity} | {(_Integer | Infinity), (_Integer | 
          Infinity)}) : {0}, opts : OptionsPattern[]] := 
  Fold[ReplaceAt[#1, rules, #2, levelspec, opts] &, expr, pos];

ReplaceAllAt[expr_, 
   rules : ((_Rule | _RuleDelayed) | {(_Rule | _RuleDelayed) ...}), 
   pos : {}] := ReplaceAll[expr, Flatten[{rules}]];

ReplaceAllAt[expr_, 
   rules : ((_Rule | _RuleDelayed) | {(_Rule | _RuleDelayed) ...}), 
   pos : ({__Integer} | _Integer)] := 
  MapAt[ReplaceAll[#, Flatten[{rules}]] &, expr, Flatten[{pos}]];

ReplaceAllAt[expr_, 
   rules : ((_Rule | _RuleDelayed) | {(_Rule | _RuleDelayed) ...}), 
   pos : {__List}] := Fold[ReplaceAllAt[#1, rules, #2] &, expr, pos];

(* The following are not used within the package 

Options[SimplifySides]=Join[{Side->Both},Options[Simplify]]

SimplifySides[expr_,opts___Rule]:=Module[{side=Side/.{opts}/.Options[SimplifySides],opts1=FilterOptions[Simplify]},
expr/.eq:Equal[x_,y_]:>If[MatchQ[side,Left|Both],Simplify[x,opts1],x]==If[MatchQ[side,Right|Both],Simplify[y,opts1],y]
]

Options[FullSimplifySides]=Join[{Side->Both},Options[Simplify]]

FullSimplifySides[expr_,opts___Rule]:=Module[{side=Side/.{opts}/.Options[FullSimplifySides],opts1=FilterOptions[FullSimplify]},
expr/.eq:Equal[x_,y_]:>If[MatchQ[side,Left|Both],FullSimplify[x,opts1],x]==If[MatchQ[side,Right|Both],FullSimplify[y,opts1],y]
]

LowestOrderTerm[expr_,var_]:=Module[{list,pos},
list=CoefficientList[expr,var];
If[list==={},0,
pos=If[Length[list]>1,Position[list,x_/;x=!=0,{1},1,Heads->False][[1,1]],1];
list[[pos]] var^(pos-1)]
]

DSolveIteration[eqns_,vars_,indvar_,param_,lastsol_]:=DSolve[eqns/.((s:a_. param^n_. #:>(s/.lastsol))&/@vars),vars,indvar][[1]];

Options[DSolveIterated]={IterationFunction->Simplify};

DSolveIterated[eqns_,vars_,indvar_,{param_,n_},opts___]:=Module[{func},
func=(MapAt[IterationFunction,#,2]&)/.{opts}/.Options[DSolveIterated];
Nest[(func/@DSolveIteration[eqns,vars,indvar,param,#])&,DSolve[eqns/.param->0,vars,indvar][[1]],n]
]

ReIm[x_]:={Re[x],Im[x]}

(*Euler angles of a vector in cartesian coordinates*)

ToEulerAngles[vec_]:={0,#[[2]],\[Pi]/2-#[[3]]}&@CoordinatesFromCartesian[vec,Spherical]

(*Matrix for rotation about an axis*)

(*AxisRotationMatrix[theta_,axis:{_,_,_}]:=Module[{n1,n2,n3},{n1,n2,n3}=axis/Sqrt[Plus@@(axis^2)];
{{n1^2+(1-n1^2) Cos[theta],n1 n2 (1-Cos[theta])+n3 Sin[theta],n1 n3 (1-Cos[theta])-n2 Sin[theta]},{n1 n2 (1-Cos[theta])-n3 Sin[theta],n2^2+(1-n2^2) Cos[theta],n2 n3 (1-Cos[theta])+n1 Sin[theta]},{n1 n3 (1-Cos[theta])+n2 Sin[theta],n2 n3 (1-Cos[theta])-n1 Sin[theta],n3^2+(1-n3^2) Cos[theta]}}]*)

(*Position of the maximum element in a list.*)

MaxPosition[table_]:=Position[table,Max[table]]

FirstSeriesTerm[expr_]:=expr/.sd:HoldPattern[SeriesData[x__]]:>If[sd[[3]]=={},0,sd[[3,1]] sd[[1]]^sd[[4]]]

Options[PerturbativeIterate]={IterationFunction->Simplify}

PerturbativeIterate[eqns_,vars_,pertvar_,lastsol_,opts___Rule]:=Module[{func},
func=IterationFunction/.{opts}/.Options[PerturbativeIterate];
MapAt[func,#,2]&/@Solve[eqns/.((s:a_. pertvar #:>(s/.lastsol))&/@vars),vars][[1]]
]

Options[PerturbativeSolve]={IterationFunction->Simplify}

PerturbativeSolve[eqns_,vars_,{pertvar_,order_},opts___Rule]:=Module[{func},
func=IterationFunction/.{opts}/.Options[PerturbativeSolve];
NestList[PerturbativeIterate[eqns,vars,pertvar,#,IterationFunction->func]&,Solve[Evaluate[eqns/.pertvar->0],vars][[1]],order]
]

PrintReturn[x_]:=(Print[x];x)

(*Generate a null m m' tensor of rank j, or a \[Kappa] q tensor decomposition *)

ZeroTriangularList[j_]:=Table[0,{k,0,2j},{q,-k,k}]

EmbedMatrix[mat1_,mat2_,pos_]:=ReplacePart[mat1,mat2,(#+pos-{1,1}&)/@#,#]&@Flatten[Array[List,Dimensions[mat2]],1]

EmbedMatrix[mat1_,mat2_]:=EmbedMatrix[mat1,mat2,{1,1}]
*)

(* ::Subsection::Closed:: *)
(*End the private context*)

End[]

(* ::Section:: *)
(*Epilog*)

(* ::Text:: *)
(*This section protects exported symbols and ends the package.*)

(* ::Subsection:: *)
(*Protect exported symbols*)

(*Protect[Evaluate[$Context<>"*"]]*)

(* ::Subsection:: *)
(*End the package context*)

EndPackage[]
