(* Mathematica Package *)

BeginPackage["AtomicDensityMatrix`CouplingCoefficients`"]
(* Exported symbols added here with SymbolName::usage *)  

CG::usage=""

TJS::usage=""

SJS::usage=""

NJS::usage=""

Begin["`Private`"] (* Begin Private Context *) 

(*Clebsch-Gordan coefficients*)

(*Triangular condition [Var88, Sec. 8.1.1, Eq. (1)]*)

CG[{F1_,m1_},{F2_,m2_},{F3_,m3_}]/;Abs[F1-F2]>F3||F1+F2<F3=0;

(*Conservation of z-projection of angular momentum [Var88, Sec. 8.1.1, Eq. (2)]*)

CG[{F1_,m1_},{F2_,m2_},{F3_,m3_}]/;m1+m2!=m3=0;

(*z-projection is less than total angular momentum [Var88, Sec. 8.1.1, condition (c)]*)

CG[{F1_,m1_},{F2_,m2_},{F3_,m3_}]/;Abs[m1]>F1||Abs[m2]>F2||Abs[m3]>F3=0;

(*CG returns the value of the Clebsch-Gordan coefficient and stores the value for future use.*)

CG[{F1_,m1_},{F2_,m2_},{F3_,m3_}]:=CG[{F1,m1},{F2,m2},{F3,m3}]=ClebschGordan[{F1,m1},{F2,m2},{F3,m3}]

(*Three-J Symbol*)

(*The conditions for the Three-J Symbol come from those for the Clebsch-Gordan coefficents.*)

(*Triangular condition*)

TJS[{F1_,m1_},{F2_,m2_},{F3_,m3_}]/;Abs[F1-F2]>F3||F1+F2<F3=0;

(*Conservation of z-projection of angular momentum*)

TJS[{F1_,m1_},{F2_,m2_},{F3_,m3_}]/;m1+m2!=-m3=0;

(*z-projection is less than total angular momentum*)

TJS[{F1_,m1_},{F2_,m2_},{F3_,m3_}]/;Abs[m1]>F1||Abs[m2]>F2||Abs[m3]>F3=0;

(*TJS returns the value of the Three-J Symbol and stores the value for future use.*)

TJS[{F1_,m1_},{F2_,m2_},{F3_,m3_}]:=TJS[{F1,m1},{F2,m2},{F3,m3}]=ThreeJSymbol[{F1,m1},{F2,m2},{F3,m3}]

(*Six-J Symbol*)

(*Triangular conditions [Var88, Sec. 9.1.1, condition (b) ]*)

SJS[{j1_,j2_,j12_},{j3_,j_,j23_}]/;Abs[j1-j2]>j12||j1+j2<j12=0;
SJS[{j1_,j2_,j12_},{j3_,j_,j23_}]/;Abs[j12-j3]>j||j12+j3<j=0;
SJS[{j1_,j2_,j12_},{j3_,j_,j23_}]/;Abs[j2-j3]>j23||j2+j3<j23=0;
SJS[{j1_,j2_,j12_},{j3_,j_,j23_}]/;Abs[j23-j1]>j||j23+j1<j=0;

(*SJS returns the value of the Six-J Symbol and stores the value for future use.*)

SJS[{j1_,j2_,j12_},{j3_,j_,j23_}]:=SJS[{j1,j2,j12},{j3,j,j23}]=SixJSymbol[{j1,j2,j12},{j3,j,j23}];

(*Nine-J Symbol*)

(*NJS returns the value of the Nine-J Symbol function [Var88, Sec. 10.2.4, Eq. (20)]*)

NJS[{a_, b_, c_}, {d_, e_, f_}, {g_, h_, j_}] := NJS[{a, b, c}, {d, e, f}, {g, h, j}] = 
Module[{xmin, xmax},
xmin = Max[{Abs[a - j], Abs[b - f], Abs[d - h]}];
xmax = Min[{Abs[a + j], Abs[b + f], Abs[d + h]}]; 
Sum[(-1)^(2 x)(2 x + 1)SJS[{a, b, c}, {f, j, x}]SJS[{d, e, f}, {b, x, h}]SJS[{g, h, j}, {x, a, d}], {x, xmin, xmax}]]

End[] (* End Private Context *)

EndPackage[]