(* Mathematica Package *)

BeginPackage["AtomicDensityMatrix`TensorAlgebra`", 
	{ "AtomicDensityMatrix`UsageMessages`",
	"AtomicDensityMatrix`Common`",
	"AtomicDensityMatrix`CouplingCoefficients`",
	"AtomicDensityMatrix`AtomicSystem`"}]

(* Exported symbols added here with SymbolName::usage *)  

StretchedState::usage=""

(*Testing and component extraction*)

ContravariantQ::usage="ContravariantQ[expr] gives True if expr is in the form of a contravariant spherical tensor or state vector, i.e., a n x 1 array (the elements of which could be operators), and gives False otherwise."

CovariantQ::usage="CovariantQ[expr] gives True if expr is in the form of a covariant spherical tensor or state vector, i.e., a 1 x n array (the elements of which could be operators), and gives False otherwise."

SphericalTensorQ::usage="SphericalTensorQ[expr] gives True if expr is in the form of a covariant or contravariant spherical tensor or state vector, i.e., a 1 x n or n x 1 array (the elements of which could be operators), and gives False otherwise."

TensorDecompositionQ::usage="TensorDecompositionQ[expr] gives True if expr is in the form of a list of spherical tensors of rank m,m+1,...,n, for some m>n, and gives False otherwise."

CovariantDecompositionQ::usage=""

ContravariantDecompositionQ::usage=""

TensorDecompositionMatrixQ::usage="TensorDecompositionMatrixQ[expr] gives True if expr is a matrix of tensor decompositions, and gives False otherwise."

CovariantDecompositionMatrixQ::usage=""

ContravariantDecompositionMatrixQ::usage=""

OperatorQ::usage="OperatorQ[expr] gives True if expr is in the form of a matrix operator, i.e., an n x n matrix, and gives False otherwise."

ScalarQ::usage="ScalarQ[expr] gives True if expr is not in the form of a Cartesian vector, spherical tensor, matrix operator or tensor decomposition, and gives False otherwise."

CartesianQ::usage="CartesianQ[expr] gives True if expr is in the form of a Cartesian vector, i.e., a one-dimensional list (the elements of which could be operators), and gives False otherwise."

OneByOneQ::usage=""

Rank::usage=""

Component::usage=""

TensorDecompositionComponent::usage=""

SphericalTensorComponent::usage=""

Components::usage=""

MapOnComponents::usage=""

(*Display*)

TensorForm::usage=""

(*Matrix multiplication*)

MatrixMultiply::usage="MatrixMultiply[a,b,c] gives inner and outer products of covariant and contravariant spherical tensors, bras and kets, operators, tensor decompositions of operators, Cartesian vectors, and scalars."

TensorNorm::usage="TensorNorm[a] gives the norm of a spherical tensor or Cartesian vector."

ContractOneByOnes::usage=""

SumRanks::usage=""

(*Conjugation*)

ExplicitConjugate::usage="ExplicitConjugate[expr] takes the complex conjugate of expr, explicitly assuming that all variables are real."

EC::usage="EC[expr] takes the complex conjugate of expr, explicitly assuming that all variables are real."

ExplicitConjugateTranspose::usage="ExplicitConjugateTranspose[expr] takes the Hermitian conjugate of a variety of types of expressions, explicitly assuming that all variables are real. If expr is a spherical tensor or matrix operator, the conjugate transpose is taken (converting covariant to contravariant, and vice versa). If expr is a tensor decomposition, the conjugate transpose of each tensor in the decomposition is taken. If expr is a Cartesian vector or a scalar, the complex conjugate is taken."

ECT::usage="ECT[expr] takes the Hermitian conjugate of a variety of types of expressions, explicitly assuming that all variables are real. If expr is a spherical tensor or matrix operator, the conjugate transpose is taken (converting covariant to contravariant, and vice versa). If expr is a tensor decomposition, the conjugate transpose of each tensor in the decomposition is taken. If expr is a Cartesian vector or a scalar, the complex conjugate is taken."

GeneralizedConjugateTranspose::usage="GeneralizedConjugateTranspose[expr] takes the Hermitian conjugate of a variety of types of expressions. If expr is a spherical tensor or matrix operator, the conjugate transpose is taken (converting covariant to contravariant, and vice versa). If expr is a tensor decomposition, the conjugate transpose of each tensor in the decomposition is taken. If expr is a Cartesian vector or a scalar, the complex conjugate is taken."

GCT::usage=""

ToCovariant::usage="ToCovariant[expr] converts any spherical tensor or Cartesian vector to a covariant spherical tensor."

ToContravariant::usage="ToContravariant[expr] converts any spherical tensor or Cartesian vector to a contravariant spherical tensor."

(*Dual*)

DualMatrix::usage=""

Dual::usage=""

TensorConjugate::usage=""

(*Spherical to Cartesian conversion*)

ToCartesian::usage="ToCartesian[expr] converts expr from a rank-one spherical tensor to a Cartesian vector. If expr is already a Cartesian vector it is left unchanged."

ToCartesianMatrix::usage=""

(*Tensor product*)

SphericalTensorProduct::usage=""

(*Error messages*)

MatrixMultiply::"coco"="Inner product attempted between two covariant or two contravariant tensors."

MatrixMultiply::"opco"="Operator multiplication attempted on wrong side of covariant or contravariant tensor."

MatrixMultiply::"sphca"="Inner product attempted between spherical and Cartesian tensors."

Begin["`Private`"] (* Begin Private Context *) 

StretchedState[j_]:=ECT@{Table[KroneckerDelta[1,i],{i,2j+1}]}

(* ::Subsubsection:: *)
(*Testing and component extraction*)

(*Check for various structures*)

(*A covariant spherical tensor is a 1xn (horizontal) array, where the elements are either all not lists, or all square matrices (operators) *)
CovariantQ[expr_]:= ArrayQ[expr]&&MatchQ[Dimensions[expr],{1,_}|{1,_,x_,x_}]

CovariantQ[expr_,test_]:=MatchQ[expr,{{__?test}}]

ContravariantQ[expr_]:=ArrayQ[expr]&&MatchQ[Dimensions[expr],{_,1}|{_,1,x_,x_}]

ContravariantQ[expr_,test_]:=MatchQ[expr,{{_?test}..}]

(* ::Input:: *)
(*ContravariantTensorsQ[A_,B_]:=(CovariantQ[A]&&ContravariantQ[B])||(CovariantQ[B]&&ContravariantQ[A])||(CartesianQ[A]&&CartesianQ[B])*)

SphericalTensorQ[expr_]:=CovariantQ[expr]||ContravariantQ[expr]

SphericalTensorQ[expr_,test_]:=CovariantQ[expr,test]||ContravariantQ[expr,test]

CartesianQ[expr_]:=ArrayQ[expr,1|3]

CartesianQ[expr_,test_]:=MatchQ[expr,{__?test}]

OperatorQ[expr_]:=MatrixQ@expr&&Equal@@Dimensions@expr

CovariantDecompositionQ[expr_]:=Head[expr]===List&&(And@@CovariantQ/@expr)&&(SphericalTensorRank/@expr)-SphericalTensorRank@First@expr===Range[Length[expr]]-1

ContravariantDecompositionQ[expr_]:=Head[expr]===List&&(And@@ContravariantQ/@expr)&&(SphericalTensorRank/@expr)-SphericalTensorRank@First@expr===Range[Length[expr]]-1

TensorDecompositionQ[expr_]:=Head[expr]===List&&(And@@SphericalTensorQ/@expr)&&(SphericalTensorRank/@expr)-SphericalTensorRank@First@expr===Range[Length[expr]]-1

TensorDecompositionMatrixQ[expr_]:=MatrixQ[expr,TensorDecompositionQ]

CovariantDecompositionMatrixQ[expr_]:=MatrixQ[expr,CovariantDecompositionQ]

ContravariantDecompositionMatrixQ[expr_]:=MatrixQ[expr,ContravariantDecompositionQ]

OneByOneQ[expr_]:=MatchQ[expr,{{_}}]

ScalarQ[expr_]:=(Head[expr]=!=List&&Head[expr]=!=Operator)

(*Returns the rank \[Kappa] of an irreducible tensor (T^\[Kappa]). For a reducible tensor T in the m m' or \[Kappa] q basis, returns the value J, where 2J is the rank of the highest irreducible tensor moment possible in the irreducible tensor decomposition of T. If T is written as a direct product of two irreducible tensors, T=A^J B^J, the irreducible tensors have rank J.*)

Rank[T_?SphericalTensorQ]:=SphericalTensorRank[T]

Rank[T_?CartesianQ]:=CartesianRank[T]

Rank[T_?OperatorQ]:=OperatorRank[T]

Rank[T_?TensorDecompositionQ]:=TensorDecompositionRank[T]

SphericalTensorRank[T_]:=(Length@SphericalTensorComponents[T]-1)/2

CartesianRank[T_]:=(Length@CartesianComponents[T]-1)/2

OperatorRank[T_]:=(Length[T]-1)/2

TensorDecompositionRank[T_]:=(Length[T]-1)/2

(*Extract or operate on all components of a vector or tensor*)

MapOnComponents[f_,expr_?CartesianQ]:=f/@expr

MapOnComponents[f_,expr_?SphericalTensorQ]:=Map[f,expr,{2}]

MapOnComponents[f_,expr_?TensorDecompositionQ]:=f/@expr

MapOnComponents[f_,expr_?OperatorQ]:=Map[f,expr,{2}]

MapOnComponents[f_,expr_?TensorDecompositionMatrixQ]:=Map[f,expr,{2}]

MapOnComponents[f_,expr_?ScalarQ]:=f[expr]

Components[T_?SphericalTensorQ]:=SphericalTensorComponents[T]

Components[T_?CartesianQ]:=CartesianComponents[T]

Components[T_?OperatorQ]:=OperatorComponents[T]

SphericalTensorComponents[T_]:=Flatten[T,1]

CartesianComponents[T_]:=T

OperatorComponents[T_]:=Flatten[T,2]

Components[system:{_AtomicState..},op_?OperatorQ,crit__Rule]:=Map[Extract[op,#]&,DirectSelfProduct[{#1,#2}&,Flatten[StatePosition[system,crit]]],{2}]

(*Individual components of vectors, tensors, or operators.*)

Component[T_?OneByOneQ]:=T[[1,1]]

Component[T_?TensorDecompositionQ,k_Integer]:=TensorDecompositionComponent[T,k]

Component[T_?TensorDecompositionQ,k_Integer,q_Integer]:=SphericalTensorComponent[TensorDecompositionComponent[T,k],q]

TensorDecompositionComponent[T_,k_Integer,q_Integer]:=SphericalTensorComponent[TensorDecompositionComponent[T,k],q]

TensorDecompositionComponent[T_,k_Integer]:=Select[T,SphericalTensorRank[#]==k&,1][[1]]

Component[T_/;OperatorQ[T]||OneByOneQ[T],m1_,m2_]:=T[[-(m1+OperatorRank[T]+1),-(m2+OperatorRank[T]+1)]]

Component[T_?SphericalTensorQ,q_/;IntegerQ[2q]]:=SphericalTensorComponent[T,q]

SphericalTensorComponent[T_,q_/;IntegerQ[2q]]:=SphericalTensorComponents[T][[-(q+SphericalTensorRank[T]+1)]]

Component[T_?CartesianQ,i_?IntegerQ]:=T[[i]]

Component[system:{_AtomicState..},op_?OperatorQ,crit__Rule]:=Components[system,op,crit][[1,1]]

Component[system:{_AtomicState..},op_?OperatorQ,{crit1__Rule},{crit2__Rule}]:=Extract[op,{First[Flatten[StatePosition[system,crit1]]],First[Flatten[StatePosition[system,crit2]]]}]

Component[sys:{_AtomicState..},op_?OperatorQ,{s1_AtomicState,s2_AtomicState}]:=op[[Position[sys,s1][[1,1]],Position[sys,s2][[1,1]]]]

Component[sys:{_AtomicState..},op_?OperatorQ,{s1_AtomicState,s2_AtomicState},kq__?IntegerQ]:=Component[Component[sys,op,{s1,s2}],kq]

Component[sys:{_AtomicState..}, op_, 
  crit1_Rule, crit2_Rule, kq__Integer] :=
 Component[Component[sys, op, {crit1}, {crit2}], kq]

Component[sys:{_AtomicState..}, 
  op_, {crit1__Rule}, {crit2__Rule}, 
  k_Integer] :=
 Component[Component[sys, op, {crit1}, {crit2}], k]

Component[sys:{_AtomicState..}, 
  op_, {crit1__Rule}, {crit2__Rule}, 
  k_Integer, q_Integer] :=
 Component[Component[sys, op, {crit1}, {crit2}], k, q]

Component[sys:{_AtomicState..}, op_, 
  crit1_Rule, crit2_Rule]/;TensorDecompositionMatrixQ[op] :=
 TensorDecompositionMatrixComponent[sys, op, {crit1}, {crit2}]

Component[sys:{_AtomicState..}, 
  op_?TensorDecompositionMatrixQ, {crit1__Rule}, {crit2__Rule}] :=
 TensorDecompositionMatrixComponent[sys, op, {crit1}, {crit2}]

TensorDecompositionMatrixComponent[sys:{_AtomicState..}, 
  op_, {crit1__Rule}, {crit2__Rule}] :=
 Module[{sys1 = RemoveZeemanSublevels@sys},
  op[[StatePosition[sys1, crit1][[1, 1]], 
    StatePosition[sys1, crit2][[1, 1]]]]
  ]

ReplaceMatrixElement[system:{_AtomicState..}, op_?OperatorQ, 
  rep : _Rule | {__Rule}, {crit1_, crit2_}] := 
 Module[{pos1 = Flatten@StatePosition[system, crit1], 
   pos2 = Flatten@StatePosition[system, crit2], matelems, op1},
  matelems = Tuples[{pos1, pos2}];
  op1 = ReplaceAllAt[op, rep, matelems];
  matelems = Reverse /@ DeleteCases[matelems, {i_, i_}];
  ReplaceAllAt[op1, rep /. (x_ -> y_) :> (x -> EC[y]), matelems]
  ]

(* ::Subsubsection:: *)
(*Display*)

TensorForm[x_?OperatorQ]:=MatrixForm[x]
TensorForm[x_?SphericalTensorQ]:=MatrixForm[x]
TensorForm[x_?TensorDecompositionQ]:=MatrixForm/@x
TensorForm[x_?TensorDecompositionMatrixQ]:=MatrixForm[Map[MatrixForm,x,{3}],TableDepth->2]
TensorForm[x_?(CartesianQ[#,ArrayQ]&)]:=MatrixForm/@x
TensorForm[x_?(CartesianQ[#,!ArrayQ[#]&]&)]:=x

(* ::Subsubsection:: *)
(*Matrix multiplication: inner and outer products*)

(*Define CenterDot to be matrix multiplication*)

LabelOperators[A_]:=A/.x_/;SphericalTensorQ[x,OperatorQ]||CartesianQ[x,OperatorQ]:>MapOnComponents[Operator,x]

Options[MatrixMultiply]={ContractOneByOnes->True,SumRanks->True}

MatrixMultiply[A_?ScalarQ,B_,opts___Rule]:=MapOnComponents[A #&,B]

MatrixMultiply[A_,B_?ScalarQ,opts___Rule]:=MapOnComponents[B #&,A]

MatrixMultiply[A_?OneByOneQ,B_?OneByOneQ,opts___Rule]:=If[ContractOneByOnes/.{opts}/.Options[MatrixMultiply],MatrixMultiply[Component[A],Component[B]],{{MatrixMultiply[Component[A],Component[B]]}}]

MatrixMultiply[A_?OperatorQ,B_?OperatorQ,opts___Rule]=A.B

MatrixMultiply[A_?TensorDecompositionQ,B_?TensorDecompositionQ,opts___Rule]:=If[SumRanks/.{opts}/.Options[MatrixMultiply],Total,Identity]@MapThread[MatrixMultiply,{A,B}]

MatrixMultiply[A_?TensorDecompositionMatrixQ,B_?TensorDecompositionMatrixQ,opts___Rule]:=If[SumRanks/.{opts}/.Options[MatrixMultiply],Total[#,2]&,Identity]@MapThread[MatrixMultiply,{A,B},2]

MatrixMultiply[A_,B_,c__,opts___Rule]:=MatrixMultiply[MatrixMultiply[A,B,opts],c,opts]

MatrixMultiply[A_,B_,opts___Rule] := (
	Inner[MatrixMultiply,LabelOperators@A,LabelOperators@B,Plus] /. Operator->Sequence // 
		If[SphericalTensorQ[A]&&SphericalTensorQ[B],Replace[#,{{x_}}->x],#]&
	) /; (CartesianQ[A]&&(CartesianQ[B]||OperatorQ[B]))||
			(CovariantQ[A]&&(ContravariantQ[B]||OperatorQ[B]))||
			(OperatorQ[A]&&(CartesianQ[B]||ContravariantQ[B]||OperatorQ[B]))||
			(ContravariantQ[A]&&CovariantQ[B])||
				Which[
					(CovariantQ[A]&&CovariantQ[B])||
					(ContravariantQ[A]&&ContravariantQ[B]),Message[MatrixMultiply::"coco"],(OperatorQ[A]&&CovariantQ[B])||
					(ContravariantQ[A]&&OperatorQ[B]),Message[MatrixMultiply::"opco"],(CartesianQ[A]&&SphericalTensorQ[B])||
					(SphericalTensorQ[A]&&CartesianQ[B]),Message[MatrixMultiply::"sphca"]
				]

MatrixMultiply[A_,B_,opts___Rule]:=
	(Inner[MatrixMultiply,LabelOperators@A,LabelOperators@B,Plus]/.Operator->Sequence//
		If[SphericalTensorQ[A]&&SphericalTensorQ[B],Replace[#,{{x_}}->x],#]&)/;
		(CartesianQ[A]&&(CartesianQ[B]||OperatorQ[B]))||
			(CovariantQ[A]&&(ContravariantQ[B]||OperatorQ[B]))||
			(OperatorQ[A]&&(CartesianQ[B]||ContravariantQ[B]||OperatorQ[B]))||
			(ContravariantQ[A]&&CovariantQ[B])||
			Which[(CovariantQ[A]&&CovariantQ[B])||(ContravariantQ[A]&&ContravariantQ[B]),
				Message[MatrixMultiply::"coco"],
				(OperatorQ[A]&&CovariantQ[B])||(ContravariantQ[A]&&OperatorQ[B]),
				Message[MatrixMultiply::"opco"],
				(CartesianQ[A]&&SphericalTensorQ[B])||(SphericalTensorQ[A]&&CartesianQ[B]),
				Message[MatrixMultiply::"sphca"]
			]

CenterDot=MatrixMultiply

System`CenterDot=CenterDot

TensorNorm[T_/;SphericalTensorQ[T]||CartesianQ[T]]:=Sqrt[ToCovariant[T]\[CenterDot]ToContravariant[T]]

TensorNorm[M_?TensorDecompositionQ]:=TensorNorm/@M

(* ::Subsubsection:: *)
(*Conjugation*)

(*"Explicit" complex conjugate, assuming all variables are real.*)

ExplicitConjugate[expr_]:=expr/.Complex[a_,b_]->Complex[a,-b]

EC=ExplicitConjugate

SuperStar=ExplicitConjugate

(*Global`SuperStar=SuperStar*)

ExplicitConjugateTranspose[T_/;SphericalTensorQ[T]||OperatorQ[T]]:=Transpose@MapOnComponents[ExplicitConjugateTranspose,T]

ExplicitConjugateTranspose[T_/;CartesianQ[T]||TensorDecompositionQ[T]||TensorDecompositionMatrixQ[T]]:=MapOnComponents[ExplicitConjugateTranspose,T]

ExplicitConjugateTranspose[T_?ScalarQ]:=EC[T]

ECT=ExplicitConjugateTranspose

GeneralizedConjugateTranspose[T_/;SphericalTensorQ[T]||OperatorQ[T]]:=Transpose@MapOnComponents[GeneralizedConjugateTranspose,T]

GeneralizedConjugateTranspose[T_/;CartesianQ[T]||TensorDecompositionQ[T]||TensorDecompositionMatrixQ[T]]:=MapOnComponents[GeneralizedConjugateTranspose,T]

GeneralizedConjugateTranspose[T_?ScalarQ]:=Conjugate[T]

GCT=GeneralizedConjugateTranspose

SuperDagger=GeneralizedConjugateTranspose

(*System`SuperDagger=GeneralizedConjugateTranspose*)

ToCovariant[T_?CovariantQ]:=T

ToCovariant[T_?ContravariantQ]:=Dual@T

ToCovariant[T_?TensorDecompositionMatrixQ]:=MapOnComponents[ToCovariant,T]

ToCovariant[T_?TensorDecompositionQ]:=ToCovariant/@T

ToContravariant[T_?ContravariantQ]:=T

ToContravariant[T_?CovariantQ]:=Dual@T

ToContravariant[T_?TensorDecompositionMatrixQ]:=MapOnComponents[ToContravariant,T]

ToContravariant[T_?TensorDecompositionQ]:=ToContravariant/@T

(* ::Subsubsection:: *)
(*Dual*)

(*Switch representation from covariant to contravariant, or vice versa. [Var88, Sec. 3.1.2, Eq. (10)]*)
(*	Subscript[A, q]=(-1)^q A^-q,	A^q=(-1)^q Subscript[A, -q]*)
(*where A is any real irreducible tensor, including real vectors. An array with dimension greater than one is assumed to be a tensor operator.*)

DualMatrix[rank_]:=Reverse@DiagonalMatrix@Reverse@Table[(-1)^q,{q,-rank,rank}]

Dual[T_?OneByOneQ]:=T

Dual[T_?TensorDecompositionMatrixQ]:=MapOnComponents[Dual,T]

Dual[T_?TensorDecompositionQ]:=Dual/@T

Dual[T_?ContravariantQ]:=Transpose[DualMatrix[Rank[T]]\[CenterDot]T]

Dual[T_?CovariantQ]:=Transpose[T\[CenterDot]DualMatrix[Rank[T]]]

Dual[T_?CartesianQ]:=T

(*Complex conjugate of an irreducible tensor [Var88, Sec. 1.2.1, Eq. (8)]*)
(*	Subscript[( *)
(*\!\(\*SuperscriptBox["A", "*"]\)), q]=*)
(*\!\(\*SuperscriptBox[*)
(*RowBox[{"(", *)
(*SuperscriptBox["A", "q"], ")"}], "*"]\)=(-1)^q *)
(*\!\(\*SuperscriptBox[*)
(*RowBox[{"(", *)
(*SubscriptBox["A", *)
(*RowBox[{"-", "q"}]], ")"}], "*"]\)*)
(*For a vector, this is equivalent to *)
(*	ToSpherical[Conjugate[ToCartesian[A]]]*)

TensorConjugate[M_?SphericalTensorQ]:=Transpose@Conjugate@Dual@M

TensorConjugate[M_?CartesianQ]:=Conjugate[M]

(* ::Subsubsection:: *)
(*Spherical to Cartesian conversion*)

(*Any vector A can be expanded in terms of basis vectors:*)
(*	A=\[Sum]A^\[Mu] Subscript[e, \[Mu]]=\[Sum]Subscript[A, \[Mu]] e^\[Mu]*)
(*Subscript[A, \[Mu]] are the covariant components*)
(*	Subscript[A, \[Mu]]=A\[CenterDot]Subscript[e, \[Mu]],*)
(*and A^\[Mu] are the contravariant components*)
(*	A^\[Mu]=A\[CenterDot]e^\[Mu].*)

(*Convert the spherical components of a vector [Var88, Sec. 1.2.1, Eq. (11)] or a set of spherical basis vectors [Var88, Sec. 1.1.5, Eq. (30)] to cartesian representation. The relations are the same in either case.*)

ToCartesianMatrix[rank_]:=Flatten[{
Table[{
Table[Switch[j,i,-1,-i,1,_,0],{j,rank,-rank,-1}],
Table[Switch[j,i,-I,-i,-I,_,0],{j,rank,-rank,-1}]
},{i,rank,1/2,-1}]/Sqrt[2],
{If[IntegerQ[rank],{Table[KroneckerDelta[j],{j,rank,-rank,-1}]},{}]}
},2]

ToCartesian[T_?ContravariantQ]:=Components@MatrixMultiply[ToCartesianMatrix@Rank@T,T]

ToCartesian[T_?CovariantQ]:=Components@MatrixMultiply[T,ECT@ToCartesianMatrix@Rank@T]

ToCartesian[T_?CartesianQ]:=T

ToContravariant[T_?CartesianQ]:=Transpose@{MatrixMultiply[ECT@ToCartesianMatrix@Rank@T,T]}

ToCovariant[T_?CartesianQ]:={T\[CenterDot]ToCartesianMatrix[Rank[T]]}

ToCartesian[sys:{_AtomicState..},op_?OperatorQ]:=Module[
	{mat=BlockDiagonal[ToCartesianMatrix/@F/@RemoveZeemanSublevels[sys]]},
	mat\[CenterDot]op\[CenterDot]ECT[mat]
]

(* ::Subsubsection:: *)
(*Tensor product*)

(*Irreducible tensor product Subscript[\!\(TraditionalForm\`{\**)
(*SubscriptBox[*)
(*StyleBox["A",*)
(*FontWeight->"Bold",*)
(*FontSlant->"Plain"], *)
(*SubscriptBox["\[Kappa]", "1"]]\[CircleTimes]\**)
(*SubscriptBox[*)
(*StyleBox["B",*)
(*FontWeight->"Bold",*)
(*FontSlant->"Plain"], *)
(*SubscriptBox["\[Kappa]", "2"]]}\), \[Kappa]] [Var88, Sec. 3.1.7, Eqs. (20,21)]*)
(*The tensor product of rank 0 is related to the scalar product by [Var88, Sec. 3.1.8, Eq. (35)]*)
(*	Subscript[A, \[Kappa]]\[CenterDot]Subscript[B, \[Kappa]]=(-1)^\[Kappa] Sqrt[2\[Kappa]+1] Subscript[{Subscript[A, \[Kappa]]\[CircleTimes]Subscript[B, \[Kappa]]}, 00]*)


SphericalTensorProduct[T1_?SphericalTensorQ,T2_?SphericalTensorQ,{k_}]:=
	{Reverse@Table[SphericalTensorProduct[T1,T2,{k,q}],{q,-k,k}]}

SphericalTensorProduct[T1_?CovariantQ,T2_?CovariantQ,{k_,q_}]:=
    Sum[
        Sum[            
            CG[{Rank[T1],q1},{Rank[T2],q2},{k,q}]
            *Component[T1,q1]
            *Component[T2,q2],
            {q2,-Rank[T2],Rank[T2]}
        ],
        {q1,-Rank[T1],Rank[T1]}
    ]
      
End[] (* End Private Context *)

EndPackage[]