(* Mathematica Package *)

Quiet[BeginPackage["AtomicDensityMatrix`WignerEckart`", { 
	"Units`", (* Units are loaded because BohrMagneton is used below. Think about updating to allow version 9 units functionality. *)
	"AtomicDensityMatrix`UsageMessages`",
	"AtomicDensityMatrix`Common`",
	"AtomicDensityMatrix`CouplingCoefficients`",
	"AtomicDensityMatrix`AtomicSystem`"}],General::obspkg];
	
(* Exported symbols added here with SymbolName::usage *)  

(* ::Subsubsection:: *)
(*Wigner-Eckart theorem*)

(*Definition*)

WignerEckart::usage=""

(* ::Subsubsection:: *)
(*Reduced matrix elements*)

ReducedME::usage=""

OperatorOptions::usage=""

EvaluateHyperfineShift::usage=""

HyperfineK::usage=""

HyperfineShift::usage=""

Identity::usage=""

Zero::usage=""

GroundState::usage=""

ExcitedState::usage=""

Polarization::usage=""

StatisticalWeight::usage=""

Dipole::usage=""

ExpandDipoleRME::usage=""

SpontaneousEmission::usage=""

LandeGJ::usage=""

LandeGF::usage=""

DipoleExpansion::usage=""

MagneticMoment::usage=""

MagneticDipole::usage=""

\[Alpha]Factor::usage=""

Begin["`Private`"] (* Begin Private Context *) 

(* ::Subsection:: *)
(*Wigner-Eckart theorem*)

(*Returns matrix elements between Zeeman sublevels in terms of reduced matrix elements. Just returns ReducedME for a toy system.*)

(*Wigner-Eckart theorem [Var88, Sec. 13.1.1, Eq. (2)]:*)
(**)
(* \[LeftAngleBracket]\[Gamma] F M|\!*)
(*\*SubsuperscriptBox[\(T\), \(q\), \(\[Kappa]\)]|\[Gamma]' F' M'\[RightAngleBracket]=(-1)^(F-M) ({*)
(* {F, \[Kappa], F'},*)
(* {-M, q, M'}*)
(*})(\[Gamma] F||T^\[Kappa]||\[Gamma]' F')*)
(* or*)
(*  \[LeftAngleBracket]\[Gamma] F M|\!*)
(*\*SubsuperscriptBox[\(T\), \(q\), \(\[Kappa]\)]|\[Gamma]' F' M'\[RightAngleBracket]=(-1)^(2q) (F'M'\[Kappa] q|F M) (\[Gamma] F||T^\[Kappa]||\[Gamma]' F')/Sqrt[2F+1]*)

Options[WignerEckart]={AtomicSystemType->Automatic,Label->None};

(*Special case for Energy*)

WignerEckart[s2_AtomicState,{Energy,0,0},s1_AtomicState,opts___Rule]:=
If[s1=!=s2,0,Switch[
ASTOpt[{opts},WignerEckart,s2],
Toy|Zeeman,
Energy[s1],
HyperfineZeeman,
Energy[s1]+HyperfineShift[J[s1],NuclearSpin[s1],F[s1],HyperfineA[s1],HyperfineB[s1]]//
	If[EvaluateHyperfineShift/.(OperatorOptions/.{opts}/.OperatorOptions->{})/.(OperatorOptions/.Options[ReducedME])/.EvaluateHyperfineShift->True,EvaluateHyperfineShift,Identity]
]
]

WignerEckart[s2_AtomicState,{Energy,0,0},s1_AtomicState,opts___Rule]:=
If[s1=!=s2,0,Switch[
ASTOpt[{opts},WignerEckart,s2],
Toy|Zeeman,
Energy[s1],
HyperfineZeeman,
Energy[s1]+HyperfineShift[J[s1],NuclearSpin[s1],F[s1],HyperfineA[s1],HyperfineB[s1]]//
	If[
		EvaluateHyperfineShift/.(OperatorOptions/.{opts}/.OperatorOptions->{})/.(OperatorOptions/.Options[ReducedME])/.EvaluateHyperfineShift->True,
		EvaluateHyperfineShift,
		Identity
	]
]
]

(*General case*)

WignerEckart[s2_AtomicState,{T_,k_,q_},s1_AtomicState,opts___Rule]:=
	Switch[
		ASTOpt[{opts},WignerEckart,s2],
		Toy,
		ReducedME[s2,{T,k},s1,Sequence@@ExpandASTOpt[{opts},WignerEckart,s2]],
		Zeeman|HyperfineZeeman,
		(2F[s2]+1)^(-1/2) CG[{F[s1],M[s1]},{k,q},{F[s2],M[s2]}]ReducedME[s2,{T,k},s1,Sequence@@ExpandASTOpt[{opts},WignerEckart,s2]]
	]

WignerEckart[s2_,{T_,k_},s1_,opts___Rule]:=Switch[
ASTOpt[{opts},WignerEckart,s2],
Toy,WignerEckart[s2,{T,k,0},s1,opts],
_,{Reverse@Table[WignerEckart[s2,{T,k,q},s1,opts],{q,-k,k}]}
]

WignerEckart[s1:{_AtomicState..},{T_,k_},opts___Rule]:=Switch[
ASTOpt[{opts},WignerEckart,s1],
Toy,{WignerEckart[s1,{T,k,0},opts]},
_,{Reverse@Table[WignerEckart[s1,{T,k,q},opts],{q,-k,k}]}
]

WignerEckart[system:{_AtomicState..},Tn_,opts___Rule]:=Module[{opts1=ExpandASTOpt[{opts},WignerEckart,system]},DirectSelfProduct[WignerEckart[#1,Tn,#2,Sequence@@opts1]&,system]]

WignerEckart[system1:{_AtomicState..},Tn_,system2:{_AtomicState..},opts___Rule]:=Module[{opts1=ExpandASTOpt[{opts},WignerEckart,system1]},DirectProduct[WignerEckart[#1,Tn,#2,Sequence@@opts1]&,system1,system2]]

WignerEckart[j_?NumericQ,Tn_,opts___Rule]:=WignerEckart[ZeemanSystem[j],Tn,AtomicSystemType->Zeeman,opts]

(* ::Subsection:: *)
(*Reduced matrix elements*)

Options[ReducedME]={AtomicSystemType->Automatic,Label->None,OperatorOptions->{}}

(* ::Subsubsection:: *)
(*Polarization operator*)

ReducedME[s1_AtomicState,{Polarization,k_},s2_AtomicState,opts___Rule]:=KroneckerDelta[StateLabel[s1],StateLabel[s2]] Sqrt[2k+1]

ReducedME[s1_AtomicState,{Polarization[{s1p_,f1_?NumericQ},{s2p_,f2_?NumericQ}],k_},s2_AtomicState,opts___Rule]:=
	KroneckerDelta[StateLabel[s1],s1p]KroneckerDelta[StateLabel[s2],s2p]KroneckerDelta[F[s1],f1]KroneckerDelta[F[s2],f2] Sqrt[2k+1]

ReducedME[s1_AtomicState,{Polarization[s1p_,s2p_],k_},s2_AtomicState,opts___Rule]:=
	KroneckerDelta[StateLabel[s1],s1p]KroneckerDelta[StateLabel[s2],s2p] Sqrt[2k+1]

(* ::Subsubsection:: *)
(*Hyperfine*)

(*Special case, statistical weight operator*)

ReducedME[s1_AtomicState,{StatisticalWeight,0},s2_AtomicState,opts___Rule]/;ASTOpt[{opts},ReducedME,s1]===HyperfineZeeman:=
	(-1)^(J[s1]+NuclearSpin[s1]+F[s2]) Sqrt[(2F[s1]+1)(2F[s2]+1)]SJS[{J[s1],F[s1],NuclearSpin[s1]},{F[s2],J[s2],0}](1/(2NuclearSpin[s1]+1) ReducedME[s1,{StatisticalWeight,0},s2,AtomicSystemType->Zeeman])

(*an angular coefficient*)

HyperfineK[J_,k_,F_]=(F+1)F-(k+1)k-(J+1)J

(*The energy of a given hyperfine level*)

EvaluateHyperfineShift[expr_]:=expr/.HyperfineShift[J_,k_,F_,HyperfineA_,HyperfineB_]:>If[J==1/2||k==1/2||J==0||k==0,1/2 HyperfineA HyperfineK[J,k,F],1/2 HyperfineA HyperfineK[J,k,F]+(3 HyperfineB*(HyperfineK[J,k,F] (HyperfineK[J,k,F]+1)-4/3 (k+1)k (J+1)J))/(8 (k (2 k-1) J*(2 J-1)))]

HyperfineShift[s_AtomicState] := EvaluateHyperfineShift@HyperfineShift[J[s],NuclearSpin[s],F[s],HyperfineA[s],HyperfineB[s]]

ReducedME[s1_AtomicState,{Energy,0},s2_AtomicState,opts___Rule] /; ASTOpt[{opts},ReducedME,s1]===HyperfineZeeman:=
	(-1)^(J[s1]+NuclearSpin[s1]+F[s2]) Sqrt[(2F[s1]+1)(2F[s2]+1)]*
	SJS[{J[s1],F[s1],NuclearSpin[s1]},{F[s2],J[s2],0}]*
	(ReducedME[s1,{Energy,0},s2,AtomicSystemType->Zeeman]+
		HyperfineShift[J[s1],NuclearSpin[s1],F[s1],HyperfineA[s1],HyperfineB[s1]] ReducedME[s1,{Identity,0},s2,AtomicSystemType->Zeeman]
	)//If[
			EvaluateHyperfineShift/.(OperatorOptions/.{opts}/.OperatorOptions->{})/.
				(OperatorOptions/.Options[ReducedME])/.EvaluateHyperfineShift->True,
			EvaluateHyperfineShift,
			Identity
		]

(*ReducedME for hyperfine structure, assuming the operator commutes with nuclear spin*)
(*(Subscript[J, 1] Subscript[F, 1]||T^(\[Kappa])||Subscript[J, 2] Subscript[F, 2])=(-1)^(Subscript[J, 1]+I+Subscript[F, 2]+\[Kappa]) Sqrt[(2Subscript[F, 1]+1)(2Subscript[F, 2]+1)]{{*)
(* {Subscript[J, 1], Subscript[F, 1], I},*)
(* {Subscript[F, 2], Subscript[J, 2], \[Kappa]}*)
(*}}(Subscript[J, 1]||T^(\[Kappa])||Subscript[J, 2])*)

ReducedME[s1_AtomicState,{T_,k_},s2_AtomicState,opts___Rule]/;ASTOpt[{opts},ReducedME,s1]===HyperfineZeeman:=(-1)^(J[s1]+NuclearSpin[s1]+F[s2]+k) Sqrt[(2F[s1]+1)(2F[s2]+1)]SJS[{J[s1],F[s1],NuclearSpin[s1]},{F[s2],J[s2],k}]ReducedME[s1,{T,k},s2,AtomicSystemType->Zeeman]

(* ::Subsubsection:: *)
(*Zero*)

ReducedME[s1_AtomicState,{Zero,0},s2_AtomicState,opts___Rule]=0;

(* ::Subsubsection:: *)
(*Identity operator*)

(*[Var88, Sec. 13.1.1, Eq. (3)]*)

ReducedME[s1_AtomicState,{Identity,0},s2_AtomicState,opts___Rule]:=KroneckerDelta[StateLabel[s1],StateLabel[s2]] Sqrt[2J[s1]+1]

(* ::Subsubsection:: *)
(*Ground state operator*)

ReducedME[s1_AtomicState,{GroundState,0},s2_AtomicState,opts___Rule]:=If[GroundStateQ[s1],ReducedME[s1,{Identity,0},s2,opts],0]

(* ::Subsubsection:: *)
(*Excited state operator*)

ReducedME[s1_AtomicState,{ExcitedState,0},s2_AtomicState,opts___Rule]:=If[!GroundStateQ[s1],ReducedME[s1,{Identity,0},s2,opts],0]

(* ::Subsubsection:: *)
(*Natural width operator*)

ReducedME[s1_AtomicState,{NaturalWidth,0},s2_AtomicState,opts___Rule]:=NaturalWidth[s1]ReducedME[s1,{Identity,0},s2,opts]

(* ::Subsubsection:: *)
(*Statistical weight operator*)

ReducedME[s1_AtomicState,{StatisticalWeight,0},s2_AtomicState,opts___Rule]:=1/(2J[s1]+1) ReducedME[s1,{Identity,0},s2,opts]

(* ::Subsubsection:: *)
(*Unperturbed energy operator*)

ReducedME[s1_AtomicState,{Energy,0},s2_AtomicState,opts___Rule]:=Energy[s1]ReducedME[s1,{Identity,0},s2,opts]

(* ::Subsubsection:: *)
(*E1 dipole operator*)

(*(\[Gamma] J||d||\[Gamma]'J')=0 for same-parity levels*)

ReducedME[s1_AtomicState,{Dipole,1},s2_AtomicState,opts___]/;Parity[s1]==Parity[s2]=0;

ReducedME[s1_AtomicState,{Dipole,1},s2_AtomicState,opts___]/;StateLabel[s1]==StateLabel[s2]=0;

(*Reorder states, using (\[Gamma] J||d||\[Gamma]'J')=(-1)^(J-J') (\[Gamma]'J'||d||\[Gamma] J) *)

ReducedME[s1_AtomicState,{Dipole,1},s2_AtomicState,opts___]/;!OrderedQ[{StateLabel[s1],StateLabel[s2]}]:=
	If[DefinedQ[s1,J],(-1)^(J[s1]-J[s2]),1]ReducedME[s2,{Dipole,1},s1,opts]

(*If states are in the right order, replace expression with a placeholder whose value can be set later*)

ReducedME[s1_AtomicState,{Dipole,1},s2_AtomicState,opts___]/;OrderedQ[{StateLabel[s1],StateLabel[s2]}]:=
	Module[
		{opts1=Label/.{opts}/.Options[ReducedME]/.None->{}},
		ReducedME[StateLabel[s1],{Dipole,1},StateLabel[s2],Sequence@@opts1]
	]

Options[ExpandDipoleRME]={AtomicSystemType->Automatic}

DipoleExpansion[s1_AtomicState,s2_AtomicState]:=
	ReducedME[StateLabel[s1],{Dipole,1},StateLabel[s2],oldopts___Rule]->
		Which[
			BranchingRatio[s2][s1]===0&&BranchingRatio[s1][s2]=!=0,
				Sqrt[3]/2 Sqrt[((1+2 J[s2])NaturalWidth[s2]BranchingRatio[s1][s2])/(Energy[s2]-Energy[s1])^3],
			BranchingRatio[s1][s2]===0&&BranchingRatio[s2][s1]=!=0,
				Sqrt[3]/2 Sqrt[((1+2 J[s1])NaturalWidth[s1]BranchingRatio[s2][s1])/(Energy[s1]-Energy[s2])^3],
			True,
				ReducedME[StateLabel[s1],{Dipole,1},StateLabel[s2],oldopts]
		]

ExpandDipoleRME[syst_,expr_,opts___Rule]:=
	expr/.Flatten@DirectSelfProduct[DipoleExpansion,Union[syst,SameTest->(StateLabel[#1]===StateLabel[#2]&)]]

(* ::Subsubsection:: *)
(*Spontaneous emission*)

ReducedME[s1_AtomicState,{SpontaneousEmission,1},s2_AtomicState,opts___Rule]/;StateLabel[s1]===StateLabel[s2]=0;

ReducedME[s1_AtomicState,{SpontaneousEmission,1},s2_AtomicState,opts___Rule]/;Parity[s1]===Parity[s2]=0;

ReducedME[s1_AtomicState,{SpontaneousEmission,1},s2_AtomicState,opts___Rule]/;!OrderedQ[{StateLabel[s1],StateLabel[s2]}]:=
	If[DefinedQ[s1,J],(-1)^(J[s1]-J[s2]),1]ReducedME[s2,{SpontaneousEmission,1},s1]

ReducedME[s1_AtomicState,{SpontaneousEmission,1},s2_AtomicState,opts___Rule]:=
	If[LevelPosition[s1]===Lower,
		Sqrt[(2J[s2]+1)NaturalWidth[s2]BranchingRatio[StateLabel[s1]][s2]],
		Sqrt[(2J[s1]+1)NaturalWidth[s1]BranchingRatio[StateLabel[s2]][s1]]
	]

(* ::Subsubsection:: *)
(*Angular momentum operator*)

ReducedME[s1_AtomicState,{J,1},s2_AtomicState,opts___Rule]:=
	Sqrt[J[s1](J[s1]+1)(2J[s1]+1)]KroneckerDelta[StateLabel[s1],StateLabel[s2]];

(* ::Subsubsection:: *)
(*Magnetic moment operator*)

(*g factors*)

LandeGJ[0,L_,S_]=0;
LandeGJ[J_,L_,S_]=1+((J+1)J+(S+1)S-(L+1)L)/(2(J+1)J);
LandeGJ[s_AtomicState]:=LandeGJ[J[s],L[s],S[s]];
SetAttributes[LandeGJ,Listable];

LandeGF[0,J_,L_,S_,k_]=0;
LandeGF[F_,J_,L_,S_,k_]=LandeGJ[J,L,S] ((F+1)F+(J+1)J-(k+1)k)/(2(F+1)F);
LandeGF[s_AtomicState]:=LandeGF[F[s],J[s],L[s],S[s],NuclearSpin[s]];
SetAttributes[LandeGF,Listable];

ReducedME[s1_AtomicState,{MagneticMoment,1},s2_AtomicState,opts___Rule]:=
	-BohrMagneton Which[
		DefinedQ[s1,GFactor], GFactor[s1],
		DefinedQ[s1,{J,L,S}], LandeGJ[s1],
		True, GFactor[s1]
		] ReducedME[s1,{J,1},s2,opts]

(* ::Subsubsection:: *)
(*MagneticDipole operator*)

(*(\[Gamma] J||d||\[Gamma]'J')=0 for same-parity levels*)

ReducedME[s1_AtomicState,{MagneticDipole,1},s2_AtomicState,opts___Rule]/;Parity[s1]!=Parity[s2]||StateLabel[s1]==StateLabel[s2]:=0

(*Reorder states, using (s1 J||MagneticDipole||\[Gamma]'J')=(-1)^(J-J') (\[Gamma]'J'||MagneticDipole||\[Gamma] J) NEED TO CHECK IF THIS IS CORRECT!*)

ReducedME[s1_AtomicState,{MagneticDipole,1},s2_AtomicState,opts___Rule]/;!OrderedQ[{StateLabel[s1],StateLabel[s2]}]:=
	(-1)^(J[s1]-J[s2]) ReducedME[s2,{MagneticDipole,1},s1]

(* ::Subsubsection:: *)
(*Polarizability operator*)

(*We define the polarizability operator of rank \[Kappa], denoted Subscript[\[Alpha], \[Kappa]], by*)
(*	Subscript[\!\(TraditionalForm\`*)
(*\*SubscriptBox[\(H\), \(Stark\)] = {\**)
(*StyleBox["E",*)
(*FontWeight->"Bold",*)
(*FontSlant->"Plain"]\[CircleTimes]\**)
(*SuperscriptBox[*)
(*StyleBox["E",*)
(*FontWeight->"Bold",*)
(*FontSlant->"Plain"], "*"]}\), \[Kappa]]\[CenterDot]Subscript[\[Alpha], \[Kappa]] ,*)
(*where Subscript[H, Stark] is effective Stark Hamiltonian and E is the (complex) electric field. In terms of the polarizabilities, the diagonal elements for a z-directed field are given by*)
(*	Subscript[\[CapitalDelta], Stark]=-(E^2/2)(Subscript[\[Alpha], 0]+Subscript[\[Alpha], 2] (3m^2-j(j+1))/j(2j-1)).*)
(*Comparison of these equations allows the determination of the reduced matrix elements of Subscript[\[Alpha], \[Kappa]].*)

\[Alpha]Factor[k_?NumericQ,j_?NumericQ]/;k>= 2j+1:=0

\[Alpha]Factor[k_?NumericQ,j_?NumericQ]:=
	\[Alpha]Factor[k,j]=Simplify[-(1/2)(Sqrt[2j+1])/(CG[{1,Mod[k,2]},{1,-Mod[k,2]},{k,0}]CG[{j,j},{k,0},{j,j}]),2j\[Element]Integers]

ReducedME[s1_AtomicState,{Polarizability,k_},s2_AtomicState,opts___Rule]:=
	\[Alpha]Factor[k,J[s1]]Polarizability[k(*,Sequence@@({Label}/.{opts}/.{Label}->{})*)][s1]KroneckerDelta[StateLabel[s1],StateLabel[s2]]

(* ::Subsubsection:: *)
(*TransitionPolarizability operator*)

ReducedME[s1_AtomicState,{TransitionPolarizability,k_},s2_AtomicState,opts___Rule]:=
	Switch[k,0,-Sqrt[3],1,Sqrt[2],2,1]TransitionPolarizability[k][s1,s2](1-KroneckerDelta[StateLabel[s1],StateLabel[s2]])

End[] (* End Private Context *)

EndPackage[]